/******************************************************************
 *
 * Copyright (c) 2003 Benjamin `Quisar' Lerman.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions  
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************/

%{

// C declarations
#include <unistd.h>
#include <stdio.h>

#define YYSTYPE int

int yyerror(char *s);
int yylex(void);
void write_number(int v);
void write_label(int v);

%}

// Bison Declaration

%token TokPUSH TokDUP TokSWAP TokDISCARD TokADD TokSUB TokMUL TokDIV TokMOD
%token TokLABEL TokCALL TokJUMP TokJUMPZ TokJUMPNEG TokENDSUB TokEND TokSTORE
%token TokRETR TokOUTCHAR TokOUTNUM TokINCHAR TokINNUM TokNUMBER TokPNUMBER

%%
//

programme:	  instruction { }
		| instruction programme { }
;

instruction:	  TokPUSH TokNUMBER { printf("  "); write_number($2); }
		| TokPUSH TokPNUMBER { printf("  "); write_number($2); }
		| TokDUP { printf(" \n "); }
		| TokSWAP { printf(" \n\t"); }
		| TokDISCARD { printf(" \n\n"); }
		| TokADD { printf("\t   "); }
		| TokSUB { printf("\t  \t"); }
		| TokMUL { printf("\t  \n"); }
		| TokDIV { printf("\t \t "); }
		| TokMOD { printf("\t \t\t"); }
		| TokLABEL TokPNUMBER  { printf("\n  "); write_label($2); }
		| TokCALL TokPNUMBER  { printf("\n \t"); write_label($2); }
		| TokJUMP TokPNUMBER  { printf("\n \n"); write_label($2); }
		| TokJUMPZ TokPNUMBER  { printf("\n\t "); write_label($2); }
		| TokJUMPNEG TokPNUMBER { printf("\n\t\t"); write_label($2); }
		| TokENDSUB  { printf("\n\t\n"); }
		| TokEND  { printf("\n\n\n"); }
		| TokSTORE  { printf("\t\t "); }
		| TokRETR  { printf("\t\t\t"); }
		| TokOUTCHAR  { printf("\t\n  "); }
		| TokOUTNUM  { printf("\t\n \t"); }
		| TokINCHAR  { printf("\t\n\t "); }
		| TokINNUM  { printf("\t\n\t\t"); }
;

%%

// Additional C code

int yyerror(char *s)
{
 fprintf(stderr, "Problme : %s\n", s);
 return 0;
}

int main()
{
  return yyparse();
}

void twrite_number(int v)
{
  if(v == 0)
    return;
  twrite_number(v/2);
  if(v%2)
    printf("\t");
  else
    printf(" ");
}

void write_number(int v)
{
  if(v >= 0)
    printf(" ");
  else
  {
    printf("\t");
    v = -v;
  }
  twrite_number(v);
  printf("\n");
}

void twrite_label(int v)
{
  if(v == 1)
    return;
  twrite_label(v/2);
  if(v%2)
    printf("\t");
  else
    printf(" ");
}

void write_label(int v)
{
  twrite_label(v);
  printf("\n");
}
