/* $Id: grammar.y,v 1.10 2002/03/10 11:25:06 quisar Exp $ */
/******************************************************************
 *
 * Copyright (c) 2003 Benjamin `Quisar' Lerman.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions  
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************/

%{

// C declarations

#include "programme.h"
#include "parser.h"

int yyerror(char *s);

%}

// Bison Declaration

%union {
 int i;
 char *s;
 value *v;
 val_list *vl;
 assign *a;
 block *b;
 block_list *bl;
 declaration *d;
 function *f;
 list_func *lf;
 memory *m;
}

%token bSTRING bCHAR bNUMBER bNAME bERROR
%token bLT bLEQ bGT bGEQ bSAME bNEQ bLAND bLOR bLNOT
%token bSEMIC bCOMMA
%token bLPAR bRPAR bLBRA bRBRA
%token bIF bELSE bWHILE bFOR bDO bASM bINT bRETURN
%token bSLASH bMINUS bMMINUS bPLUS bPPLUS bNOT bAND bOR bPCT bSTAR bCIRC
%token bASGN bASSLASH bASMINUS bASPLUS bASLSH bASRSH bASAND bASOR bASPCT
%token bASSTAR bASCIRC
%token bLBRK bRBRK bQUEST bCOLON bRSH bLSH bEXIT
%token bGETINT bGETCHAR bOUTINT bOUTCHAR bMALLOC

%left bQUEST
%left bCIRC bLSH bRSH bLT bLEQ bGT bGEQ bSAME bNEQ bAND bOR
%left bLOR
%left bLAND
%left bMINUS bPLUS
%left bSTAR bSLASH bPCT
%left bUNARY bLNOT bNOT

%type <i> bNUMBER bCHAR
%type <s> bSTRING bNAME
%type <v> val trueval
%type <vl> val_list rval_list
%type <a> assign
%type <b> block
%type <bl> list_block valorassign_list rvalorassign_list
%type <d> declaration list_name rlist_name
%type <f> func
%type <lf> list_func
%type <m> mem mem2

%%
//

programme:	declaration list_func { myprg = new programme($1, $2); }
;

list_func:	  { $$ = new list_func(); }
		| list_func func { $1->push_back($2); $$=$1; }
;

func:		bNAME bLPAR list_name bRPAR bLBRA declaration list_block bRBRA { $$ = new function($1, $3, $6, $7); }
;

list_name:	  { $$ = new declaration(); }
		| rlist_name { $$ = $1; }
;

rlist_name:	  bNAME { $$ = new declaration(); $$->push_back($1); }
		| rlist_name bCOMMA bNAME { $1->push_back($3); $$ = $1; }
;

list_block:	  { $$ = new block_list(); }
		| list_block block { $1->push_back($2); $$ = $1; }
;

block:	  trueval bSEMIC { $$ = new block_val($1); }
	| assign bSEMIC { $$ = new block_assign($1); }
	| bIF bLPAR val bRPAR block { $$ = new block_if($3, $5, new block_nop()); }
	| bIF bLPAR val bRPAR block bELSE block { $$ = new block_if($3, $5, $7); }
	| bWHILE bLPAR val bRPAR block { $$ = new block_while($3, $5); }
	| bFOR bLPAR valorassign_list bSEMIC val bSEMIC valorassign_list bRPAR block { $$ = new block_for($3, $5, $7, $9); }
	| bLBRA list_block bRBRA { $$ = new block_block_list($2); }
	| bDO block bWHILE bLPAR val bRPAR bSEMIC { $$ = new block_do($2,$5); }
	| bASM bLPAR bSTRING bRPAR bSEMIC { $$ = new block_asm($3); }
	| bRETURN val bSEMIC { $$ = new block_return($2); }
	| bEXIT val bSEMIC { $$ = new block_exit($2); }
	| bOUTINT val bSEMIC { $$ = new block_outint($2); }
	| bOUTCHAR val bSEMIC { $$ = new block_outchar($2); }
	| bSEMIC { $$ = new block_nop() ; }
;

val:	  trueval { $$ = $1; }
	| bSTAR val %prec bUNARY { $$ = new dereference($2); }
;

trueval:  bNAME { $$ = new var($1); }
	| bNUMBER { $$ = new number($1); }
	| bCHAR { $$ = new number($1); }
	| bSTRING { $$ = new valstring($1); };
	| bNAME bMMINUS { $$ = new varmm($1); } /* x-- */
	| bNAME bPPLUS { $$ = new varpp($1); } /* x++ */
	| bMMINUS bNAME { $$ = new mmvar($2); } /* --x */
	| bPPLUS bNAME { $$ = new ppvar($2); } /* ++x */
	| bMINUS val %prec bUNARY { $$ = new mval($2); }
	| bPLUS val %prec bUNARY { $$ = $2; }
	| val bSLASH val { $$ = new slash($1, $3); }
	| val bMINUS val { $$ = new myminus($1, $3); }
	| val bPLUS val { $$ = new myplus($1, $3); }
	| val bSTAR val { $$ = new star($1, $3); }
	| val bPCT val { $$ = new pct($1, $3); }
	| val bCIRC val { $$ = new circ($1, $3); }
	| val bLSH val { $$ = new lsh($1, $3); }
	| val bRSH val { $$ = new rsh($1, $3); }
	| bNOT val %prec bUNARY { $$ = new notval($2); }
	| val bAND val { $$ = new myand($1, $3); }
	| val bOR val { $$ = new myor($1, $3); }
	| bLPAR val bRPAR { $$ = $2; }
	| val bLT val { $$ = new lt($1, $3); }
	| val bLEQ val { $$ = new leq($1, $3); }
	| val bGT val { $$ = new gt($1, $3); }
	| val bGEQ val { $$ = new geq($1, $3); }
	| val bSAME val { $$ = new same($1, $3); }
	| val bNEQ val { $$ = new neq($1, $3); }
	| val bLAND val { $$ = new land($1, $3); }
	| val bLOR val { $$ = new lor($1, $3); }
	| bLNOT val %prec bUNARY { $$ = new lnotval($2); }
	| bAND bNAME %prec bUNARY { $$ = new reference($2); }
	| val bQUEST val bCOLON val { $$ = new quest($1, $3, $5); }
	| bNAME bLPAR val_list bRPAR { $$ = new appfunc($1, $3); }
	| bMALLOC bLPAR val bRPAR { $$ = new mallocfunc($3); }
;

val_list:	  { $$ = new val_list(); }
		| rval_list { $$ = $1; }
;

rval_list:	  val { $$ = new val_list(); $$->push_back($1); }
		| rval_list bCOMMA val { $1->push_back($3); $$ = $1; }
;

assign:	  mem bASGN val { $$ = new assign_val($1, $3); }
	| mem bASSLASH val { $$ = new assign_slash($1, $3); }
	| mem bASMINUS val { $$ = new assign_minus($1, $3); }
	| mem bASPLUS val { $$ = new assign_plus($1, $3); }
	| mem bASLSH val { $$ = new assign_lsh($1, $3); }
	| mem bASRSH val { $$ = new assign_rsh($1, $3); }
	| mem bASAND val { $$ = new assign_and($1, $3); }
	| mem bASOR val { $$ = new assign_or($1, $3); }
	| mem bASPCT val { $$ = new assign_pct($1, $3); }
	| mem bASSTAR val { $$ = new assign_star($1, $3); }
	| mem bASCIRC val { $$ = new assign_circ($1, $3); }
	| bGETINT mem { $$ = new assign_getint($2); }
	| bGETCHAR mem { $$ = new assign_getchar($2); }
;

mem:	  bNAME { $$ = new memory(0, $1, 0); }
	| bSTAR mem2 { $$ = $2; }
//	| bSTAR trueval { $$ = new memory(0, "", $2); }
//	| bSTAR mem  { $$ = new memory($2, "", 0); }
;

mem2:     trueval { $$ = new memory(0, "", $1); }
	| bSTAR mem2 { $$ = new memory($2, "", 0); }
;

valorassign_list:	  { $$ = new block_list(); }
			| rvalorassign_list { $$ = $1; }
;

rvalorassign_list:	  trueval { $$ = new block_list(); $$->push_back(new block_val($1)); }
			| assign { $$ = new block_list(); $$->push_back(new block_assign($1));}
			| rvalorassign_list bCOMMA trueval { $1->push_back(new block_val($3)); $$ = $1; }
			| rvalorassign_list bCOMMA assign { $1->push_back(new block_assign($3)); $$ = $1; }
;

declaration:	  { $$ = new declaration; }
		| declaration bINT bNAME bSEMIC { $1->push_back($3); $$=$1; }
;

%%

// Additional C code

int yyerror(char *s)
{
 fprintf(stderr, "Problme : %s\n", s);
 return 0;
}
