/* $Id: parser.c,v 1.6 2002/03/06 19:18:06 quisar Exp $ */
/******************************************************************
 *
 * Copyright (c) 2003 Benjamin `Quisar' Lerman.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions  
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

extern "C" {

#include <mem.h>
#include <cpp.h>

}

#include "programme.h"
#include "parser.h"

#include "grammar.tab.h"

struct lexer_state ls;

int init_lexer (void)
{
  int i;

  /* step 1 */
  init_cpp ();

  /* step 2 */
  no_special_macros = 0;
  emit_defines = emit_assertions = 0;

  /* step 3 -- with assertions */
  init_tables (1);

  /* step 4 -- no default include path */
  init_include_path (0);

  /* step 5 -- no need to reset the two emit_* variables set in 2 */
  emit_dependencies = 0;

  /* step 6 -- we work with stdin, this is not a real filename */
  set_init_filename ("[stdin]", 0);

  /* step 7 -- we make sure that assertions are on, and pragma are
     handled */
  init_lexer_state (&ls);
  init_lexer_mode (&ls);
  ls.flags |= HANDLE_ASSERTIONS | HANDLE_PRAGMA | LINE_NUM;

  /* step 8 -- input is from stdin */
  ls.input = stdin;

  /* step 10 -- we are a lexer and we want CONTEXT tokens */
  enter_file (&ls, ls.flags);

  return 0;
}

int yylex ()
{
  int r;

  /* read tokens until end-of-input is reached -- errors (non-zero
     return values different from CPPERR_EOF) are ignored */
  while ((r = lex (&ls)) < CPPERR_EOF)
  {
    if (r)
    {
      /* error condition -- no token was retrieved */
      continue;
    }
    switch (ls.ctok->type)
    {
      case NONE:
      case NEWLINE:
      case COMMENT:
      case BUNCH:
      case PRAGMA:
      case CONTEXT:
	break;

      case SLASH:
	return bSLASH;

      case ASSLASH:
	return bASSLASH;

      case MINUS:
	return bMINUS;

      case MMINUS:
	return bMMINUS;

      case ASMINUS:
	return bASMINUS;

      case PLUS:
	return bPLUS;

      case PPLUS:
	return bPPLUS;

      case ASPLUS:
	return bASPLUS;

      case LT:
	return bLT;

      case LEQ:
	return bLEQ;

      case LSH:
	return bLSH;

      case ASLSH:
	return bASLSH;

      case GT:
	return bGT;

      case GEQ:
	return bGEQ;

      case RSH:
	return bRSH;

      case ASRSH:
	return bASRSH;

      case ASGN:
	return bASGN;

      case SAME:
	return bSAME;

      case NOT:
	return bNOT;

      case NEQ:
	return bNEQ;

      case AND:
	return bAND;

      case LAND:
	return bLAND;

      case ASAND:
	return bASAND;

      case OR:
	return bOR;

      case LOR:
	return bLOR;

      case ASOR:
	return bASOR;

      case PCT:
	return bPCT;

      case ASPCT:
	return bASPCT;

      case STAR:
	return bSTAR;

      case ASSTAR:
	return bASSTAR;

      case CIRC:
	return bCIRC;

      case ASCIRC:
	return bASCIRC;

      case LNOT:
	return bLNOT;

      case LBRA:
	return bLBRA;

      case RBRA:
	return bRBRA;

      case LBRK:
	return bLBRK;

      case RBRK:
	return bRBRK;

      case LPAR:
	return bLPAR;

      case RPAR:
	return bRPAR;

      case COMMA:
	return bCOMMA;

      case QUEST:
	return bQUEST;

      case SEMIC:
	return bSEMIC;

      case COLON:
	return bCOLON;

      case NUMBER:
	yylval.i = atoi (ls.ctok->name);
	return bNUMBER;

      case STRING:
	yylval.s = convert_string (ls.ctok->name);
	return bSTRING;

      case CHAR:
	yylval.i = convert_char (ls.ctok->name);
	return bCHAR;

      case NAME:
	return scanstring (ls.ctok->name);


      default:
	return bERROR;
    }
  }
  return 0;
}

char *convert_string (char *s)
{
  char *res;
  char *tmp;
  res = (char *) malloc (strlen (s) - 1);
  if (res == NULL)
    return res;
  tmp = res;
  s++;
  while (*s && *s != '"')
  {
    if (*s == '\\')
    {
      s++;
      switch (*s)
      {
	case 'a':
	  *tmp = '\a';
	  break;

	case 'b':
	  *tmp = '\b';
	  break;

	case 'f':
	  *tmp = '\f';
	  break;

	case 'n':
	  *tmp = '\n';
	  break;

	case 'r':
	  *tmp = '\r';
	  break;

	case 't':
	  *tmp = '\t';
	  break;

	case 'v':
	  *tmp = '\v';
	  break;

	default:
	  *tmp = *s;
      }
    }
    else
      *tmp = *s;
    tmp++;
    s++;
  }
  *tmp = '\0';
  res = (char *) realloc (res, strlen (res) + 1);
  return res;
}

char convert_char (char *s)
{
  char res;
  s++;
  if(*s!='\\')
    return *s;
  else
  {
    s++;
    switch(*s)
    {
      case 'a':
	return('\a');

      case 'b':
	return('\b');

      case 'f':
	return('\f');

      case 'n':
	return('\n');

      case 'r':
	return('\n');

      case 't':
	return('\t');

      case 'v':
	return('\v');

      default:
	return(*s);
    }
  }
}
