/* $Id: programme.cpp,v 1.4 2002/03/10 21:42:07 quisar Exp $ */
/******************************************************************
 *
 * Copyright (c) 2003 Benjamin `Quisar' Lerman.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions  
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************/

#include "programme.h"

int nb_labels=37;
std::map<std::string,int> label_func;
std::map<char *,int> adress_strings;

int var::print_code(FILE *f, environnement *env)
{
  std::string comm;
  
  if(env->get_var(comm, name)) return 1;
  fprintf(f, comm.c_str());
  fprintf(f, "RETR\n");
  return 0;
}

int number::print_code(FILE *f, environnement *env)
{
  fprintf(f, "PUSH %d\n", val);
  return 0;
}

int valstring::print_code(FILE *f, environnement *env)
{
  fprintf(f, "PUSH %d\n", adress_strings[str]);
  return 0;
}

int varmm::print_code(FILE *f, environnement *env)
{
  std::string comm;
  
  if(env->get_var(comm, name)) return 1;
  fprintf(f, comm.c_str());
  fprintf(f, "DUP\nRETR\nSWAP\nDUP\nRETR\nPUSH 1\nSUB\nSTORE\n");
  return 0;
}

int mmvar::print_code(FILE *f, environnement *env)
{
  std::string comm;
  
  if(env->get_var(comm, name)) return 1;
  fprintf(f, comm.c_str());
  fprintf(f, "DUP\nDUP\nRETR\nPUSH 1\nSUB\nSTORE\nRETR\n");
  return 0;
}

int varpp::print_code(FILE *f, environnement *env)
{
  std::string comm;
  
  if(env->get_var(comm, name)) return 1;
  fprintf(f, comm.c_str());
  fprintf(f, "DUP\nRETR\nSWAP\nDUP\nRETR\nPUSH 1\nADD\nSTORE\n");
  return 0;
}

int ppvar::print_code(FILE *f, environnement *env)
{
  std::string comm;
  
  if(env->get_var(comm, name)) return 1;
  fprintf(f, comm.c_str());
  fprintf(f, "DUP\nDUP\nRETR\nPUSH 1\nADD\nSTORE\nRETR\n");
  return 0;
}

int mval::print_code(FILE *f, environnement *env)
{
  if(val->print_code(f,env)) return 1;
  fprintf(f, "PUSH 0\nSWAP\nSUB\n");
  return 0;
}

int notval::print_code(FILE *f, environnement *env) /*done in lib num 1 */
{
  fprintf(f, "PUSH -1\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "CALL 1\nSWAP\nPUSH -1\nSWAP\nSTORE\n");
  boolnotval = true;
  return 0;
}

int lnotval::print_code(FILE *f, environnement *env)
{
  int label;

  label = get_new_label();

  if(val->print_code(f,env)) return 1;
  fprintf(f, "PUSH 1\nSWAP\nJUMPZ %d\nDISCARD\nPUSH 0\nLABEL %d\n", label, label);
  return 0;
}

int dereference::print_code(FILE *f, environnement *env)
{
  if(val->print_code(f,env)) return 1;
  fprintf(f, "RETR\n");
  return 0;
}

int reference::print_code(FILE *f, environnement *env)
{
  std::string comm;

  if(env->get_var(comm, name)) return 1;
  fprintf(f, comm.c_str());
  return 0;
}

int slash::print_code(FILE *f, environnement *env)
{
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "DIV\n");
  return 0;
}

int myminus::print_code(FILE *f, environnement *env)
{
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "SUB\n");
  return 0;
}

int myplus::print_code(FILE *f, environnement *env)
{
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "ADD\n");
  return 0;
}

int star::print_code(FILE *f, environnement *env)
{
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "MUL\n");
  return 0;
}

int pct::print_code(FILE *f, environnement *env)
{
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "MOD\n");
  return 0;
}

int circ::print_code(FILE *f, environnement *env) /*done in lib num 2 */
{
  fprintf(f, "PUSH -1\nRETR\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "CALL 2\nSWAP\nPUSH -1\nSWAP\nSTORE\n");
  boolcirc = true;
  return 0;
}

int lsh::print_code(FILE *f, environnement *env) /*done in lib num 7 */
{
  fprintf(f, "PUSH -1\nRETR\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "CALL 7\nSWAP\nPUSH -1\nSWAP\nSTORE\n");
  boollsh = true;
  return 0;
}

int rsh::print_code(FILE *f, environnement *env) /*done in lib num 8 */
{
  fprintf(f, "PUSH -1\nRETR\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "CALL 8\nSWAP\nPUSH -1\nSWAP\nSTORE\n");
  boolrsh = true;
  return 0;
}

int myand::print_code(FILE *f, environnement *env) /*done in lib num 4*/
{
  fprintf(f, "PUSH -1\nRETR\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "CALL 4\nSWAP\nPUSH -1\nSWAP\nSTORE\n");
  booland = true;
  return 0;
}

int myor::print_code(FILE *f, environnement *env) /*done in lib num 5*/
{
  fprintf(f, "PUSH -1\nRETR\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "CALL 5\nSWAP\nPUSH -1\nSWAP\nSTORE\n");
  boolor = true;
  return 0;
}

int lt::print_code(FILE *f, environnement *env)
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH 1\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "SUB\nJUMPNEG %d\nDISCARD\nPUSH 0\nLABEL %d\n", label, label);
  return 0;
}

int leq::print_code(FILE *f, environnement *env)
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH 1\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "SUB\nDUP\nJUMPNEG %d\nDUP\nJUMPZ %d\nDISCARD\nPUSH 0\nSWAP\nLABEL %d\nDISCARD\n", label, label,label);
  return 0;
}

int gt::print_code(FILE *f, environnement *env)
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH 1\n");
  if(right->print_code(f,env)) return 1;
  if(left->print_code(f,env)) return 1;
  fprintf(f, "SUB\nJUMPNEG %d\nDISCARD\nPUSH 0\nLABEL %d\n", label, label);
  return 0;
}

int geq::print_code(FILE *f, environnement *env)
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH 1\n");
  if(right->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  if(left->print_code(f,env)) return 1;
  fprintf(f, "SUB\nDUP\nJUMPNEG %d\nDUP\nJUMPZ %d\nDISCARD\nPUSH 0\nSWAP\nLABEL %d\nDISCARD\n", label, label,label);
  return 0;
}

int same::print_code(FILE *f, environnement *env)
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH 1\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "SUB\nJUMPZ %d\nDISCARD\nPUSH 0\nLABEL %d\n", label, label);
  return 0;
}

int neq::print_code(FILE *f, environnement *env)
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH 0\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "SUB\nJUMPZ %d\nDISCARD\nPUSH 1\nLABEL %d\n", label, label);
  return 0;
}

int land::print_code(FILE *f, environnement *env) /*done in lib num 3*/
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH -1\nRETR\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "CALL 3\nSWAP\nPUSH -1\nSWAP\nSTORE\n"); 
  boolland = true;
  return 0;
}

int lor::print_code(FILE *f, environnement *env) /*done in lib num 6*/
{
  int label;

  label = get_new_label();

  fprintf(f, "PUSH -1\nRETR\n");
  if(left->print_code(f,env)) return 1;
  if(right->print_code(f,env)) return 1;
  fprintf(f, "CALL 6\nSWAP\nPUSH -1\nSWAP\nSTORE\n");
  boollor = true;
  return 0;
}

int quest::print_code(FILE *f, environnement *env)
{
  int labelno,labelend;
  
  labelno = get_new_label();
  labelend = get_new_label();
  
  if(test->print_code(f,env)) return 1;
  fprintf(f, "JUMPZ %d\n", labelno);
  if(iftrue->print_code(f,env)) return 1;
  fprintf(f, "JUMP %d\nLABEL %d\n", labelend, labelno);
  if(iffalse->print_code(f,env)) return 1;
  fprintf(f, "LABEL %d\n", labelend);
  return 0;
}

int appfunc::print_code(FILE *f, environnement *env)
{
  int i,ret;

#if 0
  ret = 1;
  for(i=0; i<myfuncs.size(); i++)
  {
    if(myfuncs[i]->name == func_name)
    {
      ret = 0;
      break;
    }
  }
  if(ret) return ret;
#endif
  
  /* On sauvegarde la valeur mem[0] sur la pile */
  
  fprintf(f, "PUSH -1\nRETR\n");
  
  /* On met les paramtres sur la pile */

  for(i=0; i<params->size(); i++)
    if(((*params)[i])->print_code(f,env)) return 1;

  /* On jmp  la fonction */

  fprintf(f, "CALL %d\n", label_func[func_name]);

  /* Quand on revient, le rsultat est a la premiere place de la pile, et
   * notre valeur 0 a la seconde donc on remet tout en place */

  fprintf(f, "SWAP\nPUSH -1\nSWAP\nSTORE\n");
  
  return 0;
}

int memory::print_code(FILE *f, environnement *env)
{
  std::string comm;
  
  if(mem)
  {
    if(mem->print_code(f,env)) return 1;
    fprintf(f, "RETR\n");
    return 0;
  }
  else if(val)
  {
    if(val->print_code(f,env)) return 1;
    return 0;
  }
  else if(name.size() > 0)
  {
    if(env->get_var(comm, name)) return 1;
    fprintf(f, comm.c_str());
    return 0;
  }
  return 1;
}

int assign_val::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  if(val->print_code(f,env)) return 1;
  fprintf(f, "STORE\n");
  return 0;
}
int assign_slash::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "DIV\nSTORE\n");
  return 0;
}
int assign_minus::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "SUB\nSTORE\n");
  return 0;
}
int assign_plus::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "ADD\nSTORE\n");
  return 0;
}
int assign_lsh::print_code(FILE *f, environnement *env) /* done in lib num 7 */
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nPUSH -1\nRETR\nSWAP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "CALL 7\nSWAP\nPUSH -1\nSWAP\nSTORE\nSTORE\n");
  boollsh = true;
  return 0;
}
int assign_rsh::print_code(FILE *f, environnement *env) /* done in lib num 8 */
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nPUSH -1\nRETR\nSWAP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "CALL 8\nSWAP\nPUSH -1\nSWAP\nSTORE\nSTORE\n");
  boolrsh = true;
  return 0;
}
int assign_and::print_code(FILE *f, environnement *env) /* done in lib num 4 */
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nPUSH -1\nRETR\nSWAP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "CALL 4\nSWAP\nPUSH -1\nSWAP\nSTORE\nSTORE\n");
  booland = true;
  return 0;
}
int assign_or::print_code(FILE *f, environnement *env) /* done in lib num 5 */
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nPUSH -1\nRETR\nSWAP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "CALL 5\nSWAP\nPUSH -1\nSWAP\nSTORE\nSTORE\n");
  boolor = true;
  return 0;
}
int assign_pct::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "MOD\nSTORE\n");
  return 0;
}
int assign_star::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "MUL\nSTORE\n");
  return 0;
}
int assign_circ::print_code(FILE *f, environnement *env) /* done in lib num 2 */
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "DUP\nPUSH -1\nRETR\nSWAP\nRETR\n");
  if(val->print_code(f,env)) return 1;
  fprintf(f, "CALL 2\nSWAP\nPUSH -1\nSWAP\nSTORE\nSTORE\n");
  boolcirc = true;
  return 0;
}
int assign_getint::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "INNUM\n");
  return 0;
}
int assign_getchar::print_code(FILE *f, environnement *env)
{
  if(mem->print_code(f,env)) return 1;
  fprintf(f, "INCHAR\n");
  return 0;
}
int block_nop::print_code(FILE *f, environnement *env)
{
  return 0;
}

int block_val::print_code(FILE *f, environnement *env)
{
  val->print_code(f,env);
  fprintf(f,"DISCARD\n");
  return 0;
}

int block_assign::print_code(FILE *f, environnement *env)
{
  ass->print_code(f,env);
  return 0;
}

int block_if::print_code(FILE *f, environnement *env)
{
  int labelno,labelend;
  
  labelno = get_new_label();
  labelend = get_new_label();
  
  if(test->print_code(f,env)) return 1;
  fprintf(f, "JUMPZ %d\n", labelno);
  if(iftrue->print_code(f,env)) return 1;
  fprintf(f, "JUMP %d\nLABEL %d\n", labelend, labelno);
  if(iffalse->print_code(f,env)) return 1;
  fprintf(f, "LABEL %d\n", labelend);
  return 0;
}

int block_while::print_code(FILE *f, environnement *env)
{
  int label1,label2;

  label1 = get_new_label();
  label2 = get_new_label();
  fprintf(f,"LABEL %d\n", label1);
  if(test->print_code(f, env)) return 1;
  fprintf(f,"JUMPZ %d\n", label2);
  if(act->print_code(f, env)) return 1;
  fprintf(f,"JUMP %d\nLABEL %d\n", label1, label2);
  return 0;
}

int block_do::print_code(FILE *f, environnement *env)
{
  int label1,label2;
  
  label1 = get_new_label();
  label2 = get_new_label();
  fprintf(f,"LABEL %d\n", label1);
  if(act->print_code(f, env)) return 1;
  if(test->print_code(f, env)) return 1;
  fprintf(f,"JUMPZ %d\nJUMP %d\nLABEL %d\n", label2, label1, label2);
  return 0;
}

int block_for::print_code(FILE *f, environnement *env)
{
  int label1,label2;
  int i;

  label1 = get_new_label();
  label2 = get_new_label();

  for(i=0; i<first->size(); i++)
    if((*first)[i]->print_code(f,env)) return 1;

  fprintf(f,"LABEL %d\n", label1);
  if(test->print_code(f, env)) return 1;
  fprintf(f,"JUMPZ %d\n",label2);
  if(act->print_code(f, env)) return 1;
  for(i=0; i<iter->size(); i++)
    if((*iter)[i]->print_code(f,env)) return 1;
  fprintf(f,"JUMP %d\nLABEL %d\n", label1, label2);
  return 0;
}

int block_asm::print_code(FILE *f, environnement *env)
{
  fprintf(f, commande.c_str());
  return 0;
}

int block_exit::print_code(FILE *f, environnement *env)
{
  if(val->print_code(f, env)) return 1;
  fprintf(f, "END\n");
  return 0;
}

int block_return::print_code(FILE *f, environnement *env)
{
  if(val->print_code(f, env)) return 1;
  fprintf(f, "ENDSUB\n");
  return 0;
}

int block_outint::print_code(FILE *f, environnement *env)
{
  if(val->print_code(f, env)) return 1;
  fprintf(f, "OUTNUM\n");
  return 0;
}

int block_outchar::print_code(FILE *f, environnement *env)
{
  if(val->print_code(f, env)) return 1;
  fprintf(f, "OUTCHAR\n");
  return 0;
}

int block_block_list::print_code(FILE *f, environnement *env)
{
  int i;
  
  for(i=0; i<lst->size(); i++)
    if((*lst)[i]->print_code(f,env)) return 1;
  return 0;
}

int function::print_code(FILE *f, environnement *env)
{
  int i,size1,size2,size;

  fprintf(f,"#function: %s\n", name.c_str());
  fprintf(f,"LABEL %d\n", label_value);
  size1 = parametres->size();
  size2 = variables->size();
  for(i=0;i<size1;i++)
  {
    fprintf(f, "PUSH -1\nRETR\nSWAP\nSTORE\nPUSH -1\nPUSH -1\nRETR\nPUSH 1\nSUB\nSTORE\n");
    env->add_variable((*parametres)[i], param, size2+i+1);
  }
  if(size2)
  {
    fprintf(f, "PUSH -1\nRETR\n");
    for(i=0; i<size2; i++)
    { 
      fprintf(f, "DUP\nPUSH 0\nSTORE\nPUSH 1\nSUB\n");
      env->add_variable((*variables)[i], local, size2 - i);
    }
    fprintf(f, "PUSH -1\nSWAP\nSTORE\n");
  }
  for(i=0; i<body->size(); i++)
    if((*body)[i]->print_code(f,env)) return 1;
  for(i=0; i<size1; i++)
    env->remove_variable((*parametres)[i]);
  for(i=0; i<size2; i++)
    env->remove_variable((*variables)[i]);
  fprintf(f, "PUSH 0\nENDSUB\n#endfunction: %s\n", name.c_str());
  return 0;
}

int mallocfunc::print_code(FILE *f, environnement *env)
{
  fprintf(f, "PUSH 0\nRETR\nDUP\n");
  if(val->print_code(f, env)) return 1;
  fprintf(f, "ADD\nPUSH 0\nSWAP\nSTORE\n");
  return 0;
}

int programme::print_code(FILE *f)
{
  environnement env;
  int i;
  std::map<char *,int>::iterator it;
  char *c;
  FILE *g;

  fprintf(f, "PUSH 0\nPUSH 1\n");
  i = 1;
  for(it = adress_strings.begin(); it != adress_strings.end(); it++)
  {
    it->second = i;
    for(c = it->first; *c; c++, i++)
      fprintf(f, "DUP\nPUSH %d\nSTORE\nPUSH 1\nADD\n", *c);
    fprintf(f, "DUP\nPUSH 0\nSTORE\nPUSH 1\nADD\n");
    i++;
  }
  fprintf(f, "STORE\n");
  for(i = 0; i < decl->size(); i++)
  {
    env.add_variable((*decl)[i], global, -2 - i);
    fprintf(f, "PUSH %d\nPUSH 0\nSTORE\n", -2 - i);
  }

  fprintf(f, "PUSH -1\nPUSH %d\nSTORE\n", -decl->size()-2);
  fprintf(f, "CALL %d\nEND\n", label_func["main"]);
  for(i = 0; i < funcs->size(); i++)
    if((*funcs)[i]->print_code(f, &env)) return 1;
  fflush(f);
  fprintf(f, "%s%s%s%s%s%s%s%s",
      boolnotval?strnotval.c_str():"",
      boolcirc?strcirc.c_str():"",
      boolland?strland.c_str():"",
      booland?strand.c_str():"",
      boolor?stror.c_str():"",
      boollor?strlor.c_str():"",
      boollsh?strlsh.c_str():"",
      boolrsh?strrsh.c_str():"");
}

void environnement::add_variable(std::string &name, var_type t, int loc)
{
  data[name].push_back(variable(name, t, loc));
}

void environnement::remove_variable(std::string &name)
{
  data[name].pop_back();
}

/* This function put the *ADDRESS* of a variable on the stack */
int environnement::get_var(std::string &commande, std::string &name)
{
  variable *v;
  char buf[255];
  
  sprintf(buf, "# Get %s\n", name.c_str());
  if(data.count(name))
  {
    v = &(data[name].back());
    switch(v->t)
    {
      case global:
	sprintf(buf + strlen(buf), "PUSH %d\n", v->location);
	commande = buf;
	return 0;
      case param:
      case local:
	sprintf(buf + strlen(buf), "PUSH -1\nRETR\nPUSH %d\nADD\n", v->location);
	commande = buf;
	return 0;
      default:
	return 1;
    }
  }
  else
    return 1;
}

int get_new_label(void)
{
  nb_labels++;
  return nb_labels - 1;
}
