/* $Id: programme.h,v 1.6 2002/03/10 11:25:06 quisar Exp $ */
/******************************************************************
 *
 * Copyright (c) 2003 Benjamin `Quisar' Lerman.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions  
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************/

#ifndef _PROGRAMMES_H_
#define _PROGRAMMES_H_

#include <map>
#include <list>
#include <cstdio>
#include <vector>
#include <string>

int get_new_label(void);

typedef std::vector<std::string> declaration;
typedef enum {local,global,param} var_type;

extern std::map<std::string,int> label_func;
extern std::map<char *,int> adress_strings;
extern std::string strnotval,strcirc,strland,strand,stror,strlor,strlsh,strrsh;
extern bool boolnotval,boolcirc,boolland,booland,boolor,boollor,boollsh,boolrsh;

class variable {
  public:
    var_type t;
    std::string name;
    int location;
    variable(std::string n, var_type nt, int loc) { name = n; t = nt; location = loc; }
};

class environnement {
  public:
    std::map<std::string,std::list<variable> >data;
    void add_variable(std::string &name, var_type t, int loc);
    void remove_variable(std::string &name);
    int get_var(std::string &commande, std::string &name);
};

class value {
  public:
    virtual int print_code(FILE *f, environnement *env)=0;
};

typedef std::vector<value *> val_list;

class var : public value {
  public:
    std::string name;

    var(std::string s) { name =s; }
    virtual int print_code(FILE *f, environnement *env);
};

class number : public value {
  public:
    int val;

    number(int v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class valstring : public value {
  public:
    char *str;

    valstring(char *s) { str = s; adress_strings[str] = 0; }
    virtual int print_code(FILE *f, environnement *env);
};

class varmm : public var {
  public:
    varmm(std::string s) : var(s) { }
    virtual int print_code(FILE *f, environnement *env);
};

class mmvar : public var {
  public:
    mmvar(std::string s) : var(s) { }
    virtual int print_code(FILE *f, environnement *env);
};

class varpp : public var {
  public:
    varpp(std::string s) : var(s) { }
    virtual int print_code(FILE *f, environnement *env);
};

class ppvar : public var {
  public:
    ppvar(std::string s) : var(s) { }
    virtual int print_code(FILE *f, environnement *env);
};

class mval : public value {
  public:
    value *val;
    mval(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class notval : public value {
  public:
    value *val;
    notval(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class lnotval : public value {
  public:
    value *val;
    lnotval(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class dereference : public value {
  public:
    value *val;
    dereference(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};
class reference : public value {
  public:
    std::string name;
    reference(std::string n) { name = n; }
    virtual int print_code(FILE *f, environnement *env);
};

class slash : public value {
  public:
    value *left;
    value *right;
    slash(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class myminus : public value {
  public:
    value *left;
    value *right;
    myminus(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class myplus : public value {
  public:
    value *left;
    value *right;
    myplus(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class star : public value {
  public:
    value *left;
    value *right;
    star(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class pct : public value {
  public:
    value *left;
    value *right;
    pct(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class circ : public value {
  public:
    value *left;
    value *right;
    circ(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class lsh : public value {
  public:
    value *left;
    value *right;
    lsh(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class rsh : public value {
  public:
    value *left;
    value *right;
    rsh(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class myand : public value { /* and et or sont dj dfinis. */
  public:
    value *left;
    value *right;
    myand(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class myor : public value {
  public:
    value *left;
    value *right;
    myor(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class lt : public value {
  public:
    value *left;
    value *right;
    lt(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class leq : public value {
  public:
    value *left;
    value *right;
    leq(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class gt : public value {
  public:
    value *left;
    value *right;
    gt(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class geq : public value {
  public:
    value *left;
    value *right;
    geq(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class same : public value {
  public:
    value *left;
    value *right;
    same(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class neq : public value {
  public:
    value *left;
    value *right;
    neq(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class land : public value {
  public:
    value *left;
    value *right;
    land(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class lor : public value {
  public:
    value *left;
    value *right;
    lor(value *l, value *r) { left = l; right = r; }
    virtual int print_code(FILE *f, environnement *env);
};

class quest : public value {
  public:
    value *test;
    value *iftrue;
    value *iffalse;
    quest(value *t, value *it, value *iff) { test = t; iftrue = it; iffalse = iff; }
    virtual int print_code(FILE *f, environnement *env);
};

class mallocfunc : public value {
  public:
    value* val;
    mallocfunc(value *v) { val=v; }
    virtual int print_code(FILE *f, environnement *env);
};

class appfunc : public value {
  public:
    std::string func_name;
    val_list *params;
    appfunc(std::string fn, val_list *p) { func_name = fn; params = p; }
    virtual int print_code(FILE *f, environnement *env);
};

class memory {
  public:
    memory *mem;
    std::string name;
    value* val;
    memory(memory *m, std::string n, value* v) { mem = m; name = n; val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class assign
{
  public:
    memory *mem;
//    std::string name;
    virtual int print_code(FILE *f, environnement *env)=0;
};

class assign_val : public assign
{
  public:
    value *val;
    assign_val(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_slash : public assign
{
  public:
    value *val;
    assign_slash(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_minus : public assign
{
  public:
    value *val;
    assign_minus(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_plus : public assign
{
  public:
    value *val;
    assign_plus(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_lsh : public assign
{
  public:
    value *val;
    assign_lsh(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_rsh : public assign
{
  public:
    value *val;
    assign_rsh(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_and : public assign
{
  public:
    value *val;
    assign_and(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_or : public assign
{
  public:
    value *val;
    assign_or(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_pct : public assign
{
  public:
    value *val;
    assign_pct(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_star : public assign
{
  public:
    value *val;
    assign_star(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_circ : public assign
{
  public:
    value *val;
    assign_circ(memory *m, value *v) { mem = m; val = v; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_getint : public assign
{
  public:
    assign_getint(memory *m) { mem = m; };
    virtual int print_code(FILE *f, environnement *env);
};

class assign_getchar : public assign
{
  public:
    assign_getchar(memory *m) { mem = m; };
    virtual int print_code(FILE *f, environnement *env);
};

class block {
  public:
    virtual int print_code(FILE *f, environnement *env)=0;
};

typedef std::vector<block *> block_list;

class block_nop : public block {
  public:
    virtual int print_code(FILE *f, environnement *env);
};

class block_val : public block {
  public:
    value *val;
    block_val(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_assign : public block {
  public:
    assign *ass;
    block_assign(assign *a) { ass = a; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_if : public block {
  public:
    value *test;
    block *iftrue;
    block *iffalse;
    block_if(value *t, block *it, block *iff) { test = t; iftrue = it; iffalse = iff; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_while : public block {
  public:
    value *test;
    block *act;
    block_while(value *t, block *a) { test = t; act = a; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_do : public block {
  public:
    block *act;
    value *test;
    block_do(block *a, value *t) { test = t; act = a; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_for : public block {
  public:
    block_list *first;
    value *test;
    block_list *iter;
    block *act;
    block_for(block_list *f, value *t, block_list *i, block *a)
      { first = f; test = t; iter = i; act = a; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_asm : public block {
  public:
    std::string commande;
    block_asm(std::string com) { commande = com; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_exit : public block {
  public:
    value *val;
    block_exit(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_return : public block {
  public:
    value *val;
    block_return(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_outint : public block {
  public:
    value *val;
    block_outint(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_outchar : public block {
  public:
    value *val;
    block_outchar(value *v) { val = v; }
    virtual int print_code(FILE *f, environnement *env);
};

class block_block_list : public block {
  public:
    block_list *lst;
    block_block_list(block_list *l) { lst = l; }
    virtual int print_code(FILE *f, environnement *env);
};

class function {
  public:
    std::string name;
    int label_value;
    bool used;
    bool written;
    declaration *parametres;
    declaration *variables;
    block_list *body;
    function(std::string n, declaration *p, declaration *v, block_list *b)
    { 
      name = n;
      parametres = p;
      variables = v;
      body = b;
      label_value = get_new_label();
      label_func[name] = label_value;
      used = false;
    }
    int print_code(FILE *f, environnement *env);
};

typedef std::vector<function *> list_func;

class programme {
  public:
    declaration *decl;
    list_func *funcs;
    programme(declaration *d, list_func *f) { decl = d; funcs = f; }
    int print_code(FILE *f);
};



extern programme *myprg;
extern list_func myfuncs;
extern int nb_labels;

#endif
