/******************************************************************
 *
 * Copyright (c) 2003 Benjamin `Quisar' Lerman.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions  
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************/

%{

#define YYSTYPE int

// C declarations
#include <unistd.h>
#include <stdio.h>

#include "machine.h"

int yyerror(char *s);
int yylex(void);

%}

// Bison Declaration

%token SPACE LF TAB

%%
//

programme:	  instruction { }
		| programme instruction { }
;

instruction:	  SPACE StackMan {  }
		| TAB SPACE Ari {  }
		| TAB TAB HeapAcc {  }
		| LF FlowCont {  }
		| TAB LF IO {  }
;

StackMan:	  SPACE number { instruction.push_back(std::pair<int,int>(PUSH, $2)); }
		| LF SPACE { instruction.push_back(std::pair<int,int>(DUP,0)); }
		| LF TAB { instruction.push_back(std::pair<int,int>(SWAP,0));  }
		| LF LF { instruction.push_back(std::pair<int,int>(DISCARD,0)); }
;

Ari:		  SPACE SPACE { instruction.push_back(std::pair<int,int>(ADD,0)); }
		| SPACE TAB { instruction.push_back(std::pair<int,int>(SUB,0)); }
		| SPACE LF { instruction.push_back(std::pair<int,int>(MUL,0)); }
		| TAB SPACE { instruction.push_back(std::pair<int,int>(INTDIV,0)); }
		| TAB TAB { instruction.push_back(std::pair<int,int>(MODULO,0)); }
;


HeapAcc:	  SPACE { instruction.push_back(std::pair<int,int>(STORE,0)); }
		| TAB { instruction.push_back(std::pair<int,int>(RETR,0));  }
;

FlowCont:	  SPACE SPACE label { labels[$3] = instruction.size(); } 
		| SPACE TAB label { instruction.push_back(std::pair<int,int>(CALL,$3)); }
		| SPACE LF label { instruction.push_back(std::pair<int,int>(JUMP,$3));  }
		| TAB SPACE label { instruction.push_back(std::pair<int,int>(JUMPZ,$3));  }
		| TAB TAB label { instruction.push_back(std::pair<int,int>(JUMPNEG,$3));  }
		| TAB LF { instruction.push_back(std::pair<int,int>(ENDSUB,0));  }
		| LF LF { instruction.push_back(std::pair<int,int>(END,0));  }
;

IO:		  SPACE SPACE { instruction.push_back(std::pair<int,int>(OUTCHAR,0)); }
		| SPACE TAB { instruction.push_back(std::pair<int,int>(OUTNUM,0));  }
		| TAB SPACE { instruction.push_back(std::pair<int,int>(INCHAR,0)); }
		| TAB TAB { instruction.push_back(std::pair<int,int>(INNUM,0));  }
;

number:		sign bits LF { $$ = $1 * $2; }
;

sign:		  SPACE { $$=1; }
		| TAB { $$=-1; }
;

bits:		  { $$=0; }
		| bits SPACE { $$ = 2*$1; }
		| bits TAB { $$ = 2*$1 + 1; }
;

label:		lbits LF { $$ = $1; }
;

lbits:		  { $$ = 1; }
		| lbits SPACE { $$ = 2*$1; }
		| lbits TAB { $$ = 2*$1 + 1; }
;

%%

// Additional C code

int yyerror(char *s)
{
  cerr << "Problem: " << s << endl;
  return 0;
}

int yylex()
{
  int c;
  do
  {
    c = getc(file);
    if(c == EOF)
      return 0;
  } while((c != '\n') && (c != ' ') && (c != '\t'));

  switch(c)
  {
    case '\n':
      return LF;
    case ' ':
      return SPACE;
    case '\t':
      return TAB;
    default:
      return 0;
  }
}
