/******************************************************************
 *
 * Copyright (c) 2003 Benjamin `Quisar' Lerman.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions  
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *   
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 ******************************************************************/

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#include "machine.h"

instruction_t instruction;
labels_t labels;

stack_t stack;
mem_t mem;

FILE *file;

int fd;

void exec_prog(int deb)
{
  mpz_class i,j;
  int k;
  unsigned char c;
  char buf[1024];
  while(1)
  {
    switch(instruction[deb].first)
    {
      case PUSH:
	stack.push_back(instruction[deb].second);
	break;
      case DUP:
	stack.push_back(stack.back());
	break;
      case SWAP:
	i = stack.back();
	stack.pop_back();
	j = stack.back();
	stack.pop_back();
	stack.push_back(i);
	stack.push_back(j);
	break;
      case DISCARD:
	stack.pop_back();
	break;
      case ADD:
	j = stack.back();
	stack.pop_back();
	i = stack.back();
	stack.pop_back();
	stack.push_back(i+j);
	break;
      case SUB:
	j = stack.back();
	stack.pop_back();
	i = stack.back();
	stack.pop_back();
	stack.push_back(i-j);
	break;
      case MUL:
	j = stack.back();
	stack.pop_back();
	i = stack.back();
	stack.pop_back();
	stack.push_back(i*j);
	break;
      case INTDIV:
	j = stack.back();
	stack.pop_back();
	i = stack.back();
	stack.pop_back();
	stack.push_back(i/j);
	break;
      case MODULO:
	j = stack.back();
	stack.pop_back();
	i = stack.back();
	stack.pop_back();
	stack.push_back(i%j);
	break;
      case CALL:
	exec_prog(labels[instruction[deb].second]);
	break;
      case JUMP:
	deb = labels[instruction[deb].second] - 1;
	break;
      case JUMPZ:
	i = stack.back();
	stack.pop_back();
	if(i==0)
	  deb = labels[instruction[deb].second] - 1;
	break;
      case JUMPNEG:
	i = stack.back();
	stack.pop_back();
	if(i<0)
	  deb = labels[instruction[deb].second] - 1;
	break;
      case ENDSUB:
	return;
      case END:
	exit(1);
      case STORE:
	i = stack.back();
	stack.pop_back();
	j = stack.back();
	stack.pop_back();
	mem[j] = i;
	break;
      case RETR:
	i = stack.back();
	stack.pop_back();
	stack.push_back(mem[i]);
	break;
      case OUTCHAR:
	c = stack.back().get_ui();
	stack.pop_back();
	cout << c;
	fflush(stdout);
	break;
      case OUTNUM:
	i = stack.back();
	stack.pop_back();
	cout << i ;
	fflush(stdout);
	break;
      case INCHAR:
	i = stack.back();
	stack.pop_back();
	if(read(fd, &c, 1)!=1)
	  mem[i] = -1;
	else
	  mem[i] = c;
	break;
      case INNUM:
	i = stack.back();
	stack.pop_back();
	k = 0;
	do
	{
	  read(fd, &c, 1);
	  if(c != '\n')
	    buf[k] = c;
	  k++;
	} while(c!='\n');
	buf[k] = '\0';
	j = atoi(buf);
	mem[i] = j;
	break;

      default:
	cerr << "I don't know what to do..." << endl;
    }
    deb++;
  }
}

void print_prog(void)
{
  unsigned int deb;
  for(deb = 0; deb < instruction.size(); deb++)
    switch(instruction[deb].first)
    {
      case PUSH:
	cout << deb << ": " << "PUSH " << instruction[deb].second << endl;
	break;
      case DUP:
	cout << deb << ": " << "DUP" << endl;
	break;
      case SWAP:
	cout << deb << ": " << "SWAP" << endl;
	break;
      case DISCARD:
	cout << deb << ": " << "DISCARD" << endl;
	break;
      case ADD:
	cout << deb << ": " << "ADD" << endl;
	break;
      case SUB:
	cout << deb << ": " << "SUB" << endl;
	break;
      case MUL:
	cout << deb << ": " << "MUL" << endl;
	break;
      case INTDIV:
	cout << deb << ": " << "INTDIV" << endl;
	break;
      case MODULO:
	cout << deb << ": " << "MODULO" << endl;
	break;
      case CALL:
	cout << deb << ": " << "CALL " << labels[instruction[deb].second] << endl;
	break;
      case JUMP:
	cout << deb << ": " << "JUMP " << labels[instruction[deb].second] << endl;
	break;
      case JUMPZ:
	cout << deb << ": " << "JUMPZ " << labels[instruction[deb].second] << endl;
	break;
      case JUMPNEG:
	cout << deb << ": " << "JUMPNEG " << labels[instruction[deb].second] << endl;
	break;
      case ENDSUB:
	cout << deb << ": " << "ENDSUB" << endl;
	break;
      case END:
	cout << deb << ": " << "END" << endl;
	break;
      case STORE:
	cout << deb << ": " << "STORE" << endl;
	break;
      case RETR:
	cout << deb << ": " << "RETR" << endl;
	break;
      case OUTCHAR:
	cout << deb << ": " << "OUTCHAR" << endl;
	break;
      case OUTNUM:
	cout << deb << ": " << "OUTNUM" << endl;
	break;
      case INCHAR:
	cout << deb << ": " << "INCHAR" << endl;
	break;
      case INNUM:
	cout << deb << ": " << "INNUM" << endl;
	break;

      default:
	cerr << "I don't know what to do..." << endl;
    }
}
